==> osm2go_0.1-1.tar.gz: osm2go-0.1/diff.c <==
/*
 * Copyright (C) 2008 Till Harbaum <till@harbaum.org>.
 *
 * This file is part of OSM2Go.
 *
 * OSM2Go is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OSM2Go is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OSM2Go.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * diff.c - generate and restore changes on the current data set
 */

#include "appdata.h"

#include <libxml/parser.h>
#include <libxml/tree.h>

#if !defined(LIBXML_TREE_ENABLED) || !defined(LIBXML_OUTPUT_ENABLED)
#error "libxml doesn't support required tree or output"
#endif

static void diff_save_tags(tag_t *tag, xmlNodePtr node) {
  while(tag) {
    xmlNodePtr tag_node = xmlNewChild(node, NULL, 
				      BAD_CAST "tag", NULL);
    xmlNewProp(tag_node, BAD_CAST "k", BAD_CAST tag->key);
    xmlNewProp(tag_node, BAD_CAST "v", BAD_CAST tag->value);
    tag = tag->next;
  }
}

static void diff_save_state_n_id(int flags, xmlNodePtr node, item_id_t id) {
  if(flags & GPX_FLAG_DELETED) 
    xmlNewProp(node, BAD_CAST "state", BAD_CAST "deleted");
  else if(flags & GPX_FLAG_NEW) 
    xmlNewProp(node, BAD_CAST "state", BAD_CAST "new");

  /* all items need an id */
  char *id_str = g_strdup_printf("%ld", id);
  xmlNewProp(node, BAD_CAST "id", BAD_CAST id_str);
  g_free(id_str);
}

static void diff_save_nodes(node_t *node, xmlNodePtr root_node) {
  /* store all modfied nodes */
  while(node) {
    if(node->flags) { 
      xmlNodePtr node_node = xmlNewChild(root_node, NULL, 
					 BAD_CAST "node", NULL);

      diff_save_state_n_id(node->flags, node_node, node->id);

      if(!(node->flags & GPX_FLAG_DELETED)) {
	char str[32];

	/* additional info is only required if the node hasn't been deleted */
	g_ascii_dtostr(str, sizeof(str), node->pos.lat);
	xmlNewProp(node_node, BAD_CAST "lat", BAD_CAST str);
	g_ascii_dtostr(str, sizeof(str), node->pos.lon);
	xmlNewProp(node_node, BAD_CAST "lon", BAD_CAST str);

	diff_save_tags(node->tag, node_node);
      }
    }
     node = node->next;
  }
}

static void diff_save_ways(way_t *way, xmlNodePtr root_node) {

  /* store all modfied ways */
  while(way) {
    if(way->flags) { 
      xmlNodePtr node_way = xmlNewChild(root_node, NULL, 
					 BAD_CAST "way", NULL);

      diff_save_state_n_id(way->flags, node_way, way->id);
      
      if(!(way->flags & GPX_FLAG_DELETED)) {
	/* additional info is only required if the way hasn't been deleted */
	node_chain_t *node_chain = way->node_chain;
	while(node_chain) {
	  xmlNodePtr node_node = xmlNewChild(node_way, NULL, 
					     BAD_CAST "nd", NULL);
	  char *id = g_strdup_printf("%ld", node_chain->node->id);
	  xmlNewProp(node_node, BAD_CAST "ref", BAD_CAST id);
	  g_free(id);
	  node_chain = node_chain->next;
	}

	diff_save_tags(way->tag, node_way);
      }
    }
 
    way = way->next;
  }
}

void diff_save(project_t *project, gpx_t *gpx) {
  gboolean clean = TRUE;
  char *diff_name = g_strdup_printf("%s/%s.diff", project->path, project->name);

  /* check if a diff is necessary */
  node_t *node = gpx->node;
  while(node && clean) {
    if(node->flags) clean = FALSE;
    node = node->next;
  }

  way_t *way = gpx->way;
  while(way && clean) {
    if(way->flags) clean = FALSE;
    way = way->next;
  }

  if(clean) {
    printf("data set is clean, removing diff if present\n");
    g_remove(diff_name);
    g_free(diff_name);
    return;
  }

  printf("data set is dirty, generating diff\n");

  LIBXML_TEST_VERSION;

  xmlDocPtr doc = xmlNewDoc(BAD_CAST "1.0");
  xmlNodePtr root_node = xmlNewNode(NULL, BAD_CAST "diff");
  xmlNewProp(root_node, BAD_CAST "name", BAD_CAST project->name);
  xmlDocSetRootElement(doc, root_node);

  diff_save_nodes(gpx->node, root_node);
  diff_save_ways(gpx->way, root_node);

  xmlSaveFormatFileEnc(diff_name, doc, "UTF-8", 1);
  xmlFreeDoc(doc);
  xmlCleanupParser();

  g_free(diff_name);
}

static int xml_get_prop_int(xmlNode *node, char *prop, int def) {
  char *str = (char*)xmlGetProp(node, BAD_CAST prop);
  int value = def;

  if(str) {
    value = strtoul(str, NULL, 10);
    xmlFree(str);
  }

  return value;
}

static int xml_get_prop_state(xmlNode *node, char *prop) {
  char *str = (char*)xmlGetProp(node, BAD_CAST prop);

  if(str) {
    if(strcasecmp(str, "new") == 0) {
      xmlFree(str);
      return GPX_FLAG_NEW;
    }

    if(strcasecmp(str, "deleted") == 0) {
      xmlFree(str);
      return GPX_FLAG_DELETED;
    }

    g_assert(0);
  }

  return GPX_FLAG_DIRTY;
}

static pos_t *xml_get_prop_pos(xmlNode *node) {
  char *str_lat = (char*)xmlGetProp(node, BAD_CAST "lat");
  char *str_lon = (char*)xmlGetProp(node, BAD_CAST "lon");

  if(!str_lon || !str_lat) {
    if(!str_lon) xmlFree(str_lon);
    if(!str_lat) xmlFree(str_lat);
    return NULL;
  }

  pos_t *pos = g_new0(pos_t, 1);
  pos->lat = g_ascii_strtod(str_lat, NULL);
  pos->lon = g_ascii_strtod(str_lon, NULL);

  xmlFree(str_lon);
  xmlFree(str_lat);

  return pos;
}

static tag_t *xml_scan_tags(xmlDoc *doc, xmlNodePtr node, gpx_t *gpx) {
  /* scan for tags */
  tag_t *first_tag = NULL;
  tag_t **tag = &first_tag;

  while(node) {
    if(node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)node->name, "tag") == 0) {
	/* attach tag to node/way */
	*tag = gpx_parse_osm_tag(gpx, doc, node);
	if(*tag) tag = &((*tag)->next);
      }
    }
    node = node->next;
  }
  return first_tag;
}


void diff_restore_node(xmlDoc *doc, xmlNodePtr node_node, gpx_t *gpx) {
  printf("Restoring node\n");

  /* read properties */
  item_id_t id = xml_get_prop_int(node_node, "id", ID_ILLEGAL);
  if(id == ID_ILLEGAL) {
    printf("  Node entry missing id\n");
    return;
  }

  int state = xml_get_prop_state(node_node, "state");
  pos_t *pos = xml_get_prop_pos(node_node);

  if(!(state & GPX_FLAG_DELETED) && !pos) {
    printf("  Node not deleted, but no valid position\n");
    return;
  }
    
  /* evaluate properties */
  node_t *node = NULL;

  switch(state) {
  case GPX_FLAG_NEW:
    printf("  Untested NEW\n");
    
    node = g_new0(node_t, 1);
    node->flags |= GPX_FLAG_NEW;
    /* TODO: setup and append to chain */
		
    g_assert(0);
    break;

  case GPX_FLAG_DELETED:
    printf("  Untested DELETE\n");
    
    if((node = gpx_get_node_by_id(gpx, id))) 
      node->flags |= GPX_FLAG_DELETED;
    
    g_assert(0);
    break;

  case GPX_FLAG_DIRTY:
    printf("  Valid id/position (DIRTY)\n");
    
    if((node = gpx_get_node_by_id(gpx, id))) {
      node->flags |= GPX_FLAG_DIRTY;
      printf("  found existing node\n");
    }
    break;

  default:
    printf("  Illegal node entry\n");
    return;
    break;
  }
  
  if(!node) {
    printf("  no valid node\n");		    
    return;
  }

  /* update id and position from diff */
  node->id = id;
  if(pos) {
    node->pos.lat = pos->lat;
    node->pos.lon = pos->lon;
    
    /* create utm coordinates */
    utm_t utm;
    utm_pos2utm(&node->pos, &utm, &gpx->bounds->center);
    node->lpos.x =  utm.easting  - gpx->bounds->center.easting;
    node->lpos.y = -utm.northing + gpx->bounds->center.northing;
    
    g_free(pos);
  }
  
  /* node may be an existing node, so remove tags to */
  /* make space for new ones */
  if(node->tag) {
    printf("  removing existing tags for diff tags\n");
    gpx_tag_free(node->tag);
    node->tag = NULL;
  }
  
  node->tag = xml_scan_tags(doc, node_node->children, gpx);
}

void diff_restore_way(xmlDoc *doc, xmlNodePtr node_node, gpx_t *gpx) {
  printf("Restoring way\n");
	      
  item_id_t id = xml_get_prop_int(node_node, "id", ID_ILLEGAL);
  if(id == ID_ILLEGAL) {
    printf("  entry missing id\n");
    return;
  }

  int state = xml_get_prop_state(node_node, "state");

  /* evaluate properties */
  way_t *way = NULL;
  switch(state) {
  case GPX_FLAG_NEW:
    printf("  Untested NEW\n");
    
    way = g_new0(way_t, 1);
    way->flags |= GPX_FLAG_NEW;
    /* TODO: setup and append to chain */
    
    g_assert(0);
    break;

  case GPX_FLAG_DELETED:
    printf("  Untested DELETE\n");
    
    if((way = gpx_get_way_by_id(gpx, id)))
      way->flags |= GPX_FLAG_DELETED;
    
    g_assert(0);
    break;

  case GPX_FLAG_DIRTY:
    printf("  Valid id (DIRTY)\n");
    
    if((way = gpx_get_way_by_id(gpx, id))) {
      way->flags |= GPX_FLAG_DIRTY;
      printf("  found existing way\n");
    }
    break;

  default:
    printf("  Illegal way entry\n");
    return;
  }

  if(!way) {
    printf("  no valid way\n");
    return;
  }
  
  /* update id from diff */
  way->id = id;
    
  /* update node_chain */
    
  /* way may be an existing way, so remove nodes to */
  /* make space for new ones */
  if(way->node_chain) {
    printf("  removing existing tags for diff tags\n");
    gpx_node_chain_free(way->node_chain);
    way->node_chain = NULL;
  }
  
  /* scan for nodes */
  node_chain_t **node_chain = &way->node_chain;
  xmlNode *nd_node = NULL;
  for(nd_node = node_node->children; nd_node; nd_node = nd_node->next) {
    if(nd_node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)nd_node->name, "nd") == 0) {
	/* attach node to node_chain */
	*node_chain = gpx_parse_osm_way_nd(gpx, doc, nd_node);
	if(*node_chain) node_chain = &((*node_chain)->next);
      }
    }
  }
  
  /* node may be an existing node, so remove tags to */
  /* make space for new ones */
  if(way->tag) {
    printf("  removing existing tags for diff tags\n");
    gpx_tag_free(way->tag);
    way->tag = NULL;
  }
  
  way->tag = xml_scan_tags(doc, node_node->children, gpx);
}

void diff_restore(appdata_t *appdata, project_t *project, gpx_t *gpx) {
  char *diff_name = g_strdup_printf("%s/%s.diff", project->path, project->name);
  
  if(!g_file_test(diff_name, G_FILE_TEST_EXISTS)) {
    printf("no diff present!\n");
    g_free(diff_name);
    return;
  }
  
  printf("diff found, applying ...\n");
  
  xmlDoc *doc = NULL;
  xmlNode *root_element = NULL;
  
  /* parse the file and get the DOM */
  if((doc = xmlReadFile(diff_name, NULL, 0)) == NULL) {
    errorf(GTK_WIDGET(appdata->window), 
	   "Error: could not parse file %s\n", diff_name);
    g_free(diff_name);
    return;
  }
  
  /* Get the root element node */
  root_element = xmlDocGetRootElement(doc);
  
  xmlNode *cur_node = NULL;
  for (cur_node = root_element; cur_node; cur_node = cur_node->next) {
    if (cur_node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)cur_node->name, "diff") == 0) {
	char *str = (char*)xmlGetProp(cur_node, BAD_CAST "name");
	if(str) {
	  printf("diff for project %s\n", str);
	  if(strcmp(project->name, str) != 0) {
	    messagef(GTK_WIDGET(appdata->window), _("Warning"), 
		     "Diff name (%s) does not match project name (%s)",
		     str, project->name);
	  }
	  xmlFree(str);
	}

	xmlNodePtr node_node = cur_node->children;
	while(node_node) {
	  if(node_node->type == XML_ELEMENT_NODE) {

	    if(strcasecmp((char*)node_node->name, "node") == 0) 
	      diff_restore_node(doc, node_node, gpx);
	    
	    if(strcasecmp((char*)node_node->name, "way") == 0) 
	      diff_restore_way(doc, node_node, gpx);
	  }
	  node_node = node_node->next;

	}
      }
    }
  }
 
  g_free(diff_name);
}
  
gboolean diff_present(project_t *project) {
  char *diff_name = g_strdup_printf("%s/%s.diff", project->path, project->name);
  
  if(!g_file_test(diff_name, G_FILE_TEST_EXISTS)) {
    printf("no diff present!\n");
    g_free(diff_name);
    return FALSE;
  }
  
  g_free(diff_name);
  return TRUE;
}

void diff_remove(project_t *project) {
  char *diff_name = g_strdup_printf("%s/%s.diff", project->path, project->name);
  g_remove(diff_name);
  g_free(diff_name);
}

==> osm2go_0.2-1.tar.gz: osm2go-0.2/diff.c <==
/*
 * Copyright (C) 2008 Till Harbaum <till@harbaum.org>.
 *
 * This file is part of OSM2Go.
 *
 * OSM2Go is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OSM2Go is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OSM2Go.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * diff.c - generate and restore changes on the current data set
 */

#include "appdata.h"

#include <libxml/parser.h>
#include <libxml/tree.h>

#if !defined(LIBXML_TREE_ENABLED) || !defined(LIBXML_OUTPUT_ENABLED)
#error "libxml doesn't support required tree or output"
#endif

static void diff_save_tags(tag_t *tag, xmlNodePtr node) {
  while(tag) {
    xmlNodePtr tag_node = xmlNewChild(node, NULL, 
				      BAD_CAST "tag", NULL);
    xmlNewProp(tag_node, BAD_CAST "k", BAD_CAST tag->key);
    xmlNewProp(tag_node, BAD_CAST "v", BAD_CAST tag->value);
    tag = tag->next;
  }
}

static void diff_save_state_n_id(int flags, xmlNodePtr node, item_id_t id) {
  if(flags & GPX_FLAG_DELETED) 
    xmlNewProp(node, BAD_CAST "state", BAD_CAST "deleted");
  else if(flags & GPX_FLAG_NEW) 
    xmlNewProp(node, BAD_CAST "state", BAD_CAST "new");

  /* all items need an id */
  char *id_str = g_strdup_printf("%ld", id);
  xmlNewProp(node, BAD_CAST "id", BAD_CAST id_str);
  g_free(id_str);
}

static void diff_save_nodes(node_t *node, xmlNodePtr root_node) {
  /* store all modfied nodes */
  while(node) {
    if(node->flags) { 
      xmlNodePtr node_node = xmlNewChild(root_node, NULL, 
					 BAD_CAST "node", NULL);

      diff_save_state_n_id(node->flags, node_node, node->id);

      if(!(node->flags & GPX_FLAG_DELETED)) {
	char str[32];

	/* additional info is only required if the node hasn't been deleted */
	g_ascii_dtostr(str, sizeof(str), node->pos.lat);
	xmlNewProp(node_node, BAD_CAST "lat", BAD_CAST str);
	g_ascii_dtostr(str, sizeof(str), node->pos.lon);
	xmlNewProp(node_node, BAD_CAST "lon", BAD_CAST str);
	snprintf(str, sizeof(str), "%ld", node->time);
	xmlNewProp(node_node, BAD_CAST "time", BAD_CAST str);

	diff_save_tags(node->tag, node_node);
      }
    }
     node = node->next;
  }
}

static void diff_save_ways(way_t *way, xmlNodePtr root_node) {

  /* store all modfied ways */
  while(way) {
    if(way->flags) { 
      xmlNodePtr node_way = xmlNewChild(root_node, NULL, 
					 BAD_CAST "way", NULL);

      diff_save_state_n_id(way->flags, node_way, way->id);
      
      if(!(way->flags & GPX_FLAG_DELETED)) {
	/* additional info is only required if the way hasn't been deleted */
	node_chain_t *node_chain = way->node_chain;
	while(node_chain) {
	  xmlNodePtr node_node = xmlNewChild(node_way, NULL, 
					     BAD_CAST "nd", NULL);
	  char *id = g_strdup_printf("%ld", node_chain->node->id);
	  xmlNewProp(node_node, BAD_CAST "ref", BAD_CAST id);
	  g_free(id);
	  node_chain = node_chain->next;
	}

	diff_save_tags(way->tag, node_way);
      }
    }
 
    way = way->next;
  }
}

void diff_save(project_t *project, gpx_t *gpx) {
  gboolean clean = TRUE;
  char *diff_name = g_strdup_printf("%s/%s.diff", project->path, project->name);

  /* check if a diff is necessary */
  node_t *node = gpx->node;
  while(node && clean) {
    if(node->flags) clean = FALSE;
    node = node->next;
  }

  way_t *way = gpx->way;
  while(way && clean) {
    if(way->flags) clean = FALSE;
    way = way->next;
  }

  if(clean) {
    printf("data set is clean, removing diff if present\n");
    g_remove(diff_name);
    g_free(diff_name);
    return;
  }

  printf("data set is dirty, generating diff\n");

  LIBXML_TEST_VERSION;

  xmlDocPtr doc = xmlNewDoc(BAD_CAST "1.0");
  xmlNodePtr root_node = xmlNewNode(NULL, BAD_CAST "diff");
  xmlNewProp(root_node, BAD_CAST "name", BAD_CAST project->name);
  xmlDocSetRootElement(doc, root_node);

  diff_save_nodes(gpx->node, root_node);
  diff_save_ways(gpx->way, root_node);

  xmlSaveFormatFileEnc(diff_name, doc, "UTF-8", 1);
  xmlFreeDoc(doc);
  xmlCleanupParser();

  g_free(diff_name);
}

static int xml_get_prop_int(xmlNode *node, char *prop, int def) {
  char *str = (char*)xmlGetProp(node, BAD_CAST prop);
  int value = def;

  if(str) {
    value = strtoul(str, NULL, 10);
    xmlFree(str);
  }

  return value;
}

static int xml_get_prop_state(xmlNode *node, char *prop) {
  char *str = (char*)xmlGetProp(node, BAD_CAST prop);

  if(str) {
    if(strcasecmp(str, "new") == 0) {
      xmlFree(str);
      return GPX_FLAG_NEW;
    }

    if(strcasecmp(str, "deleted") == 0) {
      xmlFree(str);
      return GPX_FLAG_DELETED;
    }

    g_assert(0);
  }

  return GPX_FLAG_DIRTY;
}

static pos_t *xml_get_prop_pos(xmlNode *node) {
  char *str_lat = (char*)xmlGetProp(node, BAD_CAST "lat");
  char *str_lon = (char*)xmlGetProp(node, BAD_CAST "lon");

  if(!str_lon || !str_lat) {
    if(!str_lon) xmlFree(str_lon);
    if(!str_lat) xmlFree(str_lat);
    return NULL;
  }

  pos_t *pos = g_new0(pos_t, 1);
  pos->lat = g_ascii_strtod(str_lat, NULL);
  pos->lon = g_ascii_strtod(str_lon, NULL);

  xmlFree(str_lon);
  xmlFree(str_lat);

  return pos;
}

static tag_t *xml_scan_tags(xmlDoc *doc, xmlNodePtr node, gpx_t *gpx) {
  /* scan for tags */
  tag_t *first_tag = NULL;
  tag_t **tag = &first_tag;

  while(node) {
    if(node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)node->name, "tag") == 0) {
	/* attach tag to node/way */
	*tag = gpx_parse_osm_tag(gpx, doc, node);
	if(*tag) tag = &((*tag)->next);
      }
    }
    node = node->next;
  }
  return first_tag;
}


void diff_restore_node(xmlDoc *doc, xmlNodePtr node_node, gpx_t *gpx) {
  printf("Restoring node\n");

  /* read properties */
  item_id_t id = xml_get_prop_int(node_node, "id", ID_ILLEGAL);
  if(id == ID_ILLEGAL) {
    printf("  Node entry missing id\n");
    return;
  }

  int state = xml_get_prop_state(node_node, "state");
  pos_t *pos = xml_get_prop_pos(node_node);

  if(!(state & GPX_FLAG_DELETED) && !pos) {
    printf("  Node not deleted, but no valid position\n");
    return;
  }
    
  /* evaluate properties */
  node_t *node = NULL;

  switch(state) {
  case GPX_FLAG_NEW:
    printf("  Restoring NEW node\n");
    
    node = g_new0(node_t, 1);
    node->visible = TRUE;
    node->flags = GPX_FLAG_NEW;
    node->time = xml_get_prop_int(node_node, "time", 0);
    if(!node->time) node->time = time(NULL);

    /* attach to end of node list */
    node_t **lnode = &gpx->node;
    while(*lnode) lnode = &(*lnode)->next;  
    *lnode = node;
    break;

  case GPX_FLAG_DELETED:
    printf("  Restoring DELETE flag\n");
    
    if((node = gpx_get_node_by_id(gpx, id))) 
      node->flags |= GPX_FLAG_DELETED;      
    else
      printf("  WARNING: no node with that id found\n");
    break;

  case GPX_FLAG_DIRTY:
    printf("  Valid id/position (DIRTY)\n");
    
    if((node = gpx_get_node_by_id(gpx, id))) 
      node->flags |= GPX_FLAG_DIRTY;
    else
      printf("  WARNING: no node with that id found\n");
    break;

  default:
    printf("  Illegal node entry\n");
    return;
    break;
  }
  
  if(!node) {
    printf("  no valid node\n");		    
    return;
  }

  /* update id and position from diff */
  node->id = id;
  if(pos) {
    node->pos.lat = pos->lat;
    node->pos.lon = pos->lon;
    
    /* create utm coordinates */
    utm_t utm;
    utm_pos2utm(&node->pos, &utm, &gpx->bounds->center);
    node->lpos.x =  utm.easting  - gpx->bounds->center.easting;
    node->lpos.y = -utm.northing + gpx->bounds->center.northing;
    
    g_free(pos);
  }
  
  /* node may be an existing node, so remove tags to */
  /* make space for new ones */
  if(node->tag) {
    printf("  removing existing tags for diff tags\n");
    gpx_tag_free(node->tag);
    node->tag = NULL;
  }
  
  node->tag = xml_scan_tags(doc, node_node->children, gpx);
}

void diff_restore_way(xmlDoc *doc, xmlNodePtr node_node, gpx_t *gpx) {
  printf("Restoring way\n");
	      
  item_id_t id = xml_get_prop_int(node_node, "id", ID_ILLEGAL);
  if(id == ID_ILLEGAL) {
    printf("  entry missing id\n");
    return;
  }

  int state = xml_get_prop_state(node_node, "state");

  /* evaluate properties */
  way_t *way = NULL;
  switch(state) {
  case GPX_FLAG_NEW:
    printf("  Untested NEW\n");
    
    way = g_new0(way_t, 1);
    way->flags |= GPX_FLAG_NEW;
    /* TODO: setup and append to chain */
    
    g_assert(0);
    break;

  case GPX_FLAG_DELETED:
    printf("  Untested DELETE\n");
    
    if((way = gpx_get_way_by_id(gpx, id)))
      way->flags |= GPX_FLAG_DELETED;
    
    g_assert(0);
    break;

  case GPX_FLAG_DIRTY:
    printf("  Valid id (DIRTY)\n");
    
    if((way = gpx_get_way_by_id(gpx, id))) {
      way->flags |= GPX_FLAG_DIRTY;
      printf("  found existing way\n");
    }
    break;

  default:
    printf("  Illegal way entry\n");
    return;
  }

  if(!way) {
    printf("  no valid way\n");
    return;
  }
  
  /* update id from diff */
  way->id = id;
    
  /* update node_chain */
    
  /* way may be an existing way, so remove nodes to */
  /* make space for new ones */
  if(way->node_chain) {
    printf("  removing existing tags for diff tags\n");
    gpx_node_chain_free(way->node_chain);
    way->node_chain = NULL;
  }
  
  /* scan for nodes */
  node_chain_t **node_chain = &way->node_chain;
  xmlNode *nd_node = NULL;
  for(nd_node = node_node->children; nd_node; nd_node = nd_node->next) {
    if(nd_node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)nd_node->name, "nd") == 0) {
	/* attach node to node_chain */
	*node_chain = gpx_parse_osm_way_nd(gpx, doc, nd_node);
	if(*node_chain) 
	  node_chain = &((*node_chain)->next);
      }
    }
  }
  
  /* node may be an existing node, so remove tags to */
  /* make space for new ones */
  if(way->tag) {
    printf("  removing existing tags for diff tags\n");
    gpx_tag_free(way->tag);
    way->tag = NULL;
  }
  
  way->tag = xml_scan_tags(doc, node_node->children, gpx);
}

void diff_restore(appdata_t *appdata, project_t *project, gpx_t *gpx) {
  char *diff_name = g_strdup_printf("%s/%s.diff", project->path, project->name);
  
  if(!g_file_test(diff_name, G_FILE_TEST_EXISTS)) {
    printf("no diff present!\n");
    g_free(diff_name);
    return;
  }
  
  printf("diff found, applying ...\n");
  
  xmlDoc *doc = NULL;
  xmlNode *root_element = NULL;
  
  /* parse the file and get the DOM */
  if((doc = xmlReadFile(diff_name, NULL, 0)) == NULL) {
    errorf(GTK_WIDGET(appdata->window), 
	   "Error: could not parse file %s\n", diff_name);
    g_free(diff_name);
    return;
  }
  
  /* Get the root element node */
  root_element = xmlDocGetRootElement(doc);
  
  xmlNode *cur_node = NULL;
  for (cur_node = root_element; cur_node; cur_node = cur_node->next) {
    if (cur_node->type == XML_ELEMENT_NODE) {
      if(strcasecmp((char*)cur_node->name, "diff") == 0) {
	char *str = (char*)xmlGetProp(cur_node, BAD_CAST "name");
	if(str) {
	  printf("diff for project %s\n", str);
	  if(strcmp(project->name, str) != 0) {
	    messagef(GTK_WIDGET(appdata->window), _("Warning"), 
		     "Diff name (%s) does not match project name (%s)",
		     str, project->name);
	  }
	  xmlFree(str);
	}

	xmlNodePtr node_node = cur_node->children;
	while(node_node) {
	  if(node_node->type == XML_ELEMENT_NODE) {

	    if(strcasecmp((char*)node_node->name, "node") == 0) 
	      diff_restore_node(doc, node_node, gpx);
	    
	    if(strcasecmp((char*)node_node->name, "way") == 0) 
	      diff_restore_way(doc, node_node, gpx);
	  }
	  node_node = node_node->next;

	}
      }
    }
  }
 
  g_free(diff_name);
}
  
gboolean diff_present(project_t *project) {
  char *diff_name = g_strdup_printf("%s/%s.diff", project->path, project->name);
  
  if(!g_file_test(diff_name, G_FILE_TEST_EXISTS)) {
    printf("no diff present!\n");
    g_free(diff_name);
    return FALSE;
  }
  
  g_free(diff_name);
  return TRUE;
}

void diff_remove(project_t *project) {
  char *diff_name = g_strdup_printf("%s/%s.diff", project->path, project->name);
  g_remove(diff_name);
  g_free(diff_name);
}
